#!/usr/bin/perl
#########################################################################
#                                                                       #
# Copyright and Intellectual Property Right Notice:                     #
#                                                                       #
# All contents of this Source Code are: Copyright  2000-2005           #
# FRISK Software International (FRISK). All rights reserved.            #
#                                                                       #
# FRISK retains all title and ownership to the Source Code and any      #
# derivative works. FRISK owns copyrights, inventions, know-how and     #
# trade secrets related to the design, operation, amendments or         #
# services of the Source Code and any derivative works which are        #
# developed from the Source Code. FRISK retains all rights to           #
# modify, prepare derivative works of, decompile, reverse engineer,     #
# and disassemble the Source Code from the Software.                    #
#                                                                       #
# Use and modification of the Source Code shall only be permitted by    #
# any licensed party pursuant a prior written authorisation from        #
# FRISK.                                                                #
#                                                                       #
#########################################################################
#
# scan-mail.pl, (c) 2003 FRISK Software International
# Henr �r Baldursson <henry@f-prot.com>
# Bjarni Rnar Einarsson <bre@f-prot.com>
#
# This script is used for virus-scanning e-mail using F-Prot Antivirus and
# the Anomy Sanitizer (http://mailtools.anomy.net/).
#
# This program can either operate as a server, a milter plugin, a postfix plugin,
# a smtp proxy or in a "one-shot" mode.
#
# In "one-shot" mode, the script receives a message on standard input and
# outputs a virus-scanned message on standard output.  This makes sense for 
# use with procmail or similar mail filtering systems.
#
# In server mode, the script communicates with other processes via. a 
# TCP socket.  Details on the protocol can be found below.
#
# In milter mode, the script will wait for connections from Sendmail's
# milter code and filter incoming messages accordingly.  Milter mode 
# requires either the F-Prot Milter plugin or the Sendmail::Milter 
# module, either the "standard" version or the pure-perl implementation
# available at http://www.duh.org/pmilter/.  Only the PMilter version 
# has been tested by the author at this time.
#
# In postfix mode, the script will read SMTP commands from STDIN, relay them
# to a postfix smtpd, whose address is specified on the commandline after the
# -postfix commandline option, and relay the output to STDOUT. Mail messages will
# be sanitized on the fly. This works so that scan-mail.pl can be run from the
# master.cf file of Postfix configuration as a service.
#
# In smtp proxy mode, scan-mail.pl will listen to a given address for SMTP
# connections and relay the connection to a SMTP daemon. Messages will be sanitized
# on the fly.
#
# The script accepts the following arguments:
#
#  -verbose                Enable some informative messages on STDERR.
#  -backup                 Enable backups, storing a copy of all processed mail in
#                          /usr/local/f-prot/backup/.
#  -quarantine             Only make backups of modified messages.
#  -server                 Enter server mode.
#  -server_port N          Listen on port N instead of the default (11200).
#  -milter [P|C]           Enter milter mode.  The optional argument is either
#                          treated as a milter connection (local:/path/to/socket/ or 
#                          inet:port@host) or the path to a sendmail.cf file to read.
#  -daemon                 Milter or server operation in the background as a daemon.
#  -postfix host:port      Enter postfix mode. The argument is an address of a smtpd
#                          scan-mail.pl should relay SMTP connections to.
#  -proxy                  Enter proxying mode.
#  -proxy_address address  When in proxying mode, listen on the address specified
#                          instead of the default (0.0.0.0:25).
#  -smtp_address address   When in proxying mode, relay messages to an MTA on the
#                          specified address instead of the default (127.0.0.1:26).
#
#
# Client/server communication overview:
#
#   1. The client connects via TCP to 127.0.0.1:11200.
#
#   2. The client requests scanning by sending a line formatted like this
#      to the server:  
#
#          /path/to/input/file<TAB>/path/to/output/file<NEWLINE>
#
#   3. The file is scanned as an e-mail messages, and a new message is
#      written to the requested location.  The server then sends a result
#      to the client, formatted like so:
#
#          /path/to/input/file: <EXIT CODE>: <MOD COUNT>: <MESSAGE><NEWLINE>
#
#   4. Restart from step 2.
#
# The exit codes are the same codes as are returned by the "one-shot" 
# invocation of the program. 
#
# Note that all filename paths MUST have all whitespace and the ":" and "%" 
# characters encoded using HTTP URL encoding (a %-sign followed by the 
# hexadecimal ascii-code of the caracter).  Other characters may optionally
# be encoded as well.  This means a space is encoded as %20, a newline as 
# %0A, a tab as %09, and so forth.
#

BEGIN { push @INC, "/usr/local/f-prot/perllib" };

# This makes things a little more secure.
umask(077);

my $debug           = 0;
my $verbose         = 0;
my $quarantine      = 0;
my $daemon          = 0;
my $milter          = 0;
my $milter_cf       = '/etc/mail/sendmail.cf';
my $milter_conn     = undef;
my $milter_name     = 'f-prot-milter';
my $milter_pid      = undef;
my $fprot_milter    = '/usr/local/f-prot/tools/f-prot-milter';
my $postfix         = 0;
my $postfix_ret     = undef;
my $server          = 0;
my $server_port     = 11200;
my $proxy           = 0;
my $proxy_addr      = '0.0.0.0:25';
my $smtp_addr       = '127.0.0.1:26';
my $backup          = 0;
my $backup_location = "/usr/local/f-prot/backup/";
my $scanmailconf = "/etc/scan-mail.pl.conf";
my $ret = 0;
my $signalflag;

use strict;
use Anomy::Sanitizer;
use Anomy::Sanitizer::FProt;
use POSIX ":sys_wait_h";

sub REAPER {
	print STDERR "Reaper in $$ called...\n" if $debug;
	1 until (waitpid(-1, WNOHANG) == -1 );
	$signalflag = 1;
	$SIG{CHLD} = \&REAPER;
}

my @SCANNER_CONF = (
    'file_list_rules = 1',
    'file_list_1_policy = accept:defang:drop:defang',
    'msg_file_drop = ***** Attachment removed by F-Prot Antivirus *****\n\nThe attachment/message\n\n%FILENAME\n\nwas a virus/worm/trojan. It was removed.\n\nThis message was generated by F-Prot Antivirus Mail Scanner.\n',
    'msg_file_save = ***** Attachment removed by F-Prot Antivirus *****\n\nThe attachment/message\n\n%FILENAME\n\nwas infected.  It could not be disinfected and was therefore\nremoved and placed in the quarantine as\n\n%SAVEDNAME\n\nThis message was generated by F-Prot Antivirus Mail Scanner.\n',
    'header_url =',
    'header_rev =',
    'feat_log_inline = 0',
    'feat_lengths = 0',
    'feat_scripts = 0',
    'feat_html = 0',
    'feat_fixmime = 0'
);

# these are hardcoded values.
my @CONF_HARDCODE = (
    'feat_files = 1',
    'file_characters =',
    'file_list_1 = .*',
    'file_list_1_scanner = 0:1:2:builtin/fprotd %FILENAME -server -disinf -dumb',
    'header_info = X-Antivirus: Scanned by F-Prot Antivirus (http://www.f-prot.com)',
    'msg_usage = Usage: cat mail | scan-mail.pl [ -backup ] [ \'variable op value\' | \'filename\' ] > virus_free_mail\n',
    'msg_defanged = ATTACHMENT_BLOCKED'
);

my $engine = new Anomy::Sanitizer;
my $fprot = new Anomy::Sanitizer::FProt;

while ($ARGV[0] =~ /^-/)
{
    my $arg = shift @ARGV;
    if (lc($arg) =~ /^--?b(ackup)?$/)
    {
        $backup = 1;
    }
    elsif (lc($arg) =~ /^--?q(uarantine)?$/)
    {
        $quarantine = 1;
    }
    elsif (lc($arg) =~ /^--?s(erver)?$/)
    {
        $server = 1;
    }
    elsif (lc($arg) =~ /^--?d(aemon)?$/)
    {
        $server = $daemon = 1;
    }
    elsif (lc($arg) =~ /^--?v(erbose)?$/)
    {
        $verbose = 1;
    }
    elsif (lc($arg) =~ /^--?(server_)?p(ort)?$/)
    {
        $server_port = shift @ARGV;
    }
    elsif (lc($arg) =~ /^--?m(ilter)?$/)
    {
        $milter = 1;
        if ($ARGV[0] =~ /^(local|unix|inet(6)?)?:/)
        {
            $milter_conn = shift @ARGV;
        }
        elsif (-e $ARGV[0])
        {
            $milter_cf = shift @ARGV;
        }
    }
    elsif (lc($arg) =~ /^--?p(ostfix)?$/)
    {
        $postfix = 1;
        $postfix_ret = shift @ARGV;
    }
    elsif (lc($arg) =~ /^--?debug$/)
    {
        $debug = 1;
    }
    elsif (lc($arg) =~ /^--?proxy$/)
    {
        $proxy = 1;
    }
    elsif (lc($arg) =~ /^--?proxy_addr(ess)?$/)
    {
        $proxy_addr = shift @ARGV;
    }
    elsif (lc($arg) =~ /^--?smtp_addr(ess)?$/)
    {
        $smtp_addr = shift @ARGV;
    }
    else
    {
        die "Invalid argument: $arg\n";
    }
}

if ($postfix)
{
    $daemon = 0 if ($daemon);
    die "Invalid commandline" if ($milter);
}

$engine->register_scanner("fprotd", $fprot->get_sanitizer_callback());
if (my $e = $engine->error())
{
    print STDERR $e;
    exit(103);
}

if (-r $scanmailconf)
{
    $engine->configure(@SCANNER_CONF, $scanmailconf, @ARGV, @CONF_HARDCODE);
}
else
{
    $engine->configure(@SCANNER_CONF, @ARGV, @CONF_HARDCODE);
}

if (my $e = $engine->error())
{
    print STDERR $engine->get_msg("usage"), "\n";
    print STDERR $engine->get_msg("current"), "\n";
    print STDERR $engine->get_config_text();
    print STDERR "\n", $e, "\n";
    exit(104);
}

# This is log magic which collects information from the F-Prot scanner 
# module so we can use it directly here.
my @infections = ( );
my @fprot_codes = ( );
$engine->{"log"}->add_hook("fprot_result", sub
{
    my ($self, $entry) = @_;
    push @fprot_codes, $entry->[3]->{code} if ($entry->[3]->{code});
    push @infections,  $entry->[3]->{name} if ($entry->[3]->{name});
    1;
}, "global");


# Daemonize?
if ($daemon)
{
    use POSIX 'setsid';
    chdir '/'                  or die "Can't chdir to /: $!";
    open STDIN, '/dev/null'    or die "Can't read /dev/null: $!";
    open STDOUT, '>/dev/null'  or die "Can't write to /dev/null: $!";
    defined(my $pid = fork)    or die "Can't fork: $!";
    exit if $pid;
    setsid                     or die "Can't start a new session: $!";
    open PID_FILE, '>/var/run/scanmail.pid' or die "Can't open pid file: $!";
    print PID_FILE $$;
    close PID_FILE;
    open STDERR, '>&STDOUT'    or die "Can't dup stdout: $!";
}

# Are we running as a milter?
if ($milter && (-x $fprot_milter))
{
    start_fprot_milter($fprot_milter);
    $server = 1;
}
elsif ($milter)
{
    eval "use ScanMilter;";
    if ($@)
    {
        print STDERR
            "ERROR:  Unsatisfied dependencies!\n",
            "\n",
            "In order to run in Milter mode, one of the following must be\n",
            "present on your system:\n",
            "  - F-Prot Antivirus's Milter: $fprot_milter\n",
            "  - The ScanMilter and Sendmail::Milter perl modules.\n",
            "Please correct these problems and try again.\n",
            "\n";
        exit(1);
    }
    
    my $args = { name     => $milter_name,
                 debug    => $verbose,
                 callback => \&milter_scan_jobs,
                 cf_path  => $milter_cf };
    $args->{conn} = $milter_conn if ($milter_conn);

    ScanMilter::Run($args);
    exit(0);
}

# Postfix spawn?
if ($postfix)
{
    use IO::Socket::INET;
    use IO::Handle;


    open my $out, ">&STDOUT" or die "Can't duplicate standard output";
    $out->autoflush();
    open my $in, "<&STDIN" or die "Can't duplicate standard input";
    $in->autoflush();

    $server = 0 if $server;

    # make a connection to other end
    print STDERR "Connecting to SMTP, " . $postfix_ret . "\n" if ($verbose);
    my $smtp = new IO::Socket::INET (  PeerAddr => $postfix_ret,
                                       Proto    => 'tcp',
                                       Type     => SOCK_STREAM
                                    );
    if (!defined $smtp)
    {
        print STDERR "Connection failed: $!\n" if ($verbose);
        exit(1);
    }
    else
    {
        $smtp->autoflush();
        print STDERR "Connected\n" if ($verbose);
    }
    handle_smtp_flow($in, $out, $smtp);
    exit(0);
}

# SMTP proxy?
if ($proxy)
{
    run_smtp_proxy($proxy_addr, $smtp_addr);
    exit(0);
}

# Are we running in server mode?
if ($server)
{
    use IO::Socket::INET;
    my $sock = new IO::Socket::INET ( LocalHost => '127.0.0.1',
                                      LocalPort => $server_port,
                                      Proto     => 'tcp',
                                      Type      => SOCK_STREAM,
                                      Listen    => 20, 
                                      Reuse     => 1 );
    die "Error creating listening socket: $!" unless ($sock);
    print STDERR "Listening on $server_port ...\n" if ($verbose);

    while (1) {
        my ($conn, $peer);
		# this is ugly, but works across all the unixes with all their signal mechanism snafu...
        while (1) { 
            ($conn, $peer) = $sock->accept();
            last if defined $conn;
        }
        my ($port, $iaddr) = sockaddr_in($peer);
        print STDERR "Accepted a connection from ", inet_ntoa($iaddr) ,"\n"
          if ($verbose);
        
        # We don't want any zombies.
        $SIG{CHLD} = \&REAPER;
        my $kid = fork();
        
        if ($kid)
        {
            $conn->close();
            next;
        }
        else
        {
            my $job;
            while( 1 ){
                $job = <$conn>;
                if(!defined $job){
                    if ( $signalflag == 1){
                        $signalflag = 0;
                        next;
                    }
                    last;
                }
                my $scanner = fork();
                next if ($scanner);
                
                my $result = scan_job($job);
                chomp $result;
                $conn->print($result, "\n");
                exit(0);
            }
            $conn->close();
            exit(0);
        }
    }
    exit(0);
}
else
{
    my ($code, $modifications, $text) = do_scan();
    chomp $text;
    print STDERR $text, "\n" if ($verbose);
    exit($code);
}


##[ Subroutines ]##############################################################

# proxy address, smtp address
sub run_smtp_proxy
{
    use IO::Socket::INET;

    my ($proxy_addr, $smtp_addr) = @_;

    my $sock = new IO::Socket::INET ( LocalHost => $proxy_addr,
                                      Proto     => 'tcp',
                                      Type      => SOCK_STREAM,
                                      Listen    => 20,
                                      Reuse     => 1 );
    die "Error creating listening socket: $!" unless ($sock);
    print STDERR "Listening for SMTP on " . $proxy_addr . "\n" if ($verbose);

    while (my ($conn, $peer) = $sock->accept())
    {
        $conn->autoflush();
        if ($verbose)
        {
            my ($port, $iaddr) = sockaddr_in($peer);
            print STDERR "Accepted a connection from ", inet_ntoa($iaddr) ,"\n";
        }
        $SIG{CHLD} = \&REAPER;
        my $kid = fork();
        if ($kid)
        {
            $conn->close();
            next;
        }
        else
        {
            # make a connection to other end
            print STDERR "Connecting to SMTP, " . $smtp_addr . "\n" if ($verbose);
            my $out = new IO::Socket::INET (  PeerAddr => $smtp_addr,
                                              Proto    => 'tcp',
                                              Type     => SOCK_STREAM
                                           );
            if (!defined $out)
            {
                print STDERR "Connection failed: $!\n" if ($verbose);
                exit(1);
            }
            else
            {
                $out->autoflush();
                print STDERR "Connected\n" if ($verbose);
            }
            open my $conn_out, ">&", $conn or die "Can't duplicate connection filehandle";
            $conn_out->autoflush();
            open my $conn_in, "<&", $conn or die "Can't duplicate connection filehandle";
            $conn_out->autoflush();
            handle_smtp_flow($conn_in, $conn_out, $out);
            exit(0);
        }
    }
}

sub handle_smtp_flow
{
    use File::Temp qw(tempfile);
    my ($in, $out, $smtp) = @_;

    my $reply = <$smtp>;
    print $out $reply;

    while ($_ = <$in>)
    {
         if (/^data(\r)?\n$/i)
         {
             my ($ofile, $ofilename) = tempfile( DIR => "/tmp" );
             printf STDERR "Output file \"%s\"\n", $ofilename if ($debug);
             close(STDOUT);
             open(STDOUT, "+>$ofilename");
             STDOUT->autoflush();
             
             my $sleep_count = 0;
             my $pid = undef;
             do {
                 $pid = open(SCAN_KID, "|-");
                 unless (defined $pid) {
                     warn "cannot fork: $!";
                     die "bailing out" if $sleep_count++ > 6;
                     sleep 2 + $sleep_count;
                 }
             } until defined $pid;
                     
             if ($pid > 0)
             {
                 print $out "354 Enter mail, end with \".\" on a line by itself\r\n";
                 # We are the parent, stream the message to the
                 # sanitizing kid process.
                 while ($_ = <$in>)
                 {
                     unless (/^\.(\r)?\n$/i)
                     {
                         # Default is to send data to kid...
                         s/^\.(\.\r?\n)$/$1/s; # un-escape '..' 
                         print SCAN_KID $_;
                     }
                     else
                     {
                         # Else, when we see the dot then close
                         # kid pipe and send the modified message 
                         # to postfix.
                         close(SCAN_KID);
                         my $exitcode = $?;
                         printf STDERR "Message complete, child exit code: %d (%d)\n", $exitcode, int($exitcode/256) 
                           if ($verbose);
                         
                         if (int($exitcode/256) != 0)
                         {
                             print $out "451 Please try again later\r\n";
                             print $smtp "RSET\r\n";
                         }
                         else
                         {
                             print $smtp "DATA\r\n";
                             $_ = <$smtp>;
                             print STDERR "Got back from DATA: " . $_ . "\n" if ($debug);
                             if (/^3\d\d\s.*/)
                             {
                                 seek(STDOUT, 0, 0);
                                 while ($_ = <STDOUT>)
                                 {
                                     if( /^\.(\r)?\n$/i ){ # escape single dot lines
                                         print $smtp ".";
                                     }
                                     print $smtp $_;
                                 }
                                 print $smtp ".\r\n";
                             }
                             else
                             {
                                 print $out "451 Please try again later!\r\n";
                             }
                         }
                         
                         close(STDOUT);
                         unlink $ofilename;
                         last;
                     }
                 }
             }
             else
             {
                 # We are the kid, clean the message and exit.
                 my ($code, $modifications, $text) = do_scan();
                 close(STDOUT);
                 print STDERR $text, "\n" if ($verbose);
                 exit($code);
             }
         }
         elsif (/^QUIT(\r)?\n$/i)
         {
             print STDERR "Connector says: " . $_ if ($debug);
             print $smtp $_;
             $_ = <$smtp>;
             print STDERR "Remote replies: " . $_ if ($debug);
             print $out $_;
             $out->close();
             $in->close();
             exit(0);
         }
         elsif (/^RCPT\s+To:\s*/i)
         {
             # if no match set a flag which disables scanning, or something
         }

         print STDERR "Connector says: " . $_ if ($debug);
         # print line to $smtp
         print $smtp $_;

         # print to $peer whatever $smtp says

         if ($smtp->eof)
         {
             print STDERR "Connection to remote lost\n" if ($verbose);
             print $out "451 Please try again later\r\n";
             exit(1); # Child dies.
         }

         while ($_ = <$smtp>)
         {
             print STDERR "Remote replies: " . $_ if ($debug);
             print $out $_;
             if (/^\d\d\d\s/)
             {
                 last;
             }
         }
    }
}

sub start_fprot_milter
{
    my $prog = shift;
    
    if (!$milter_conn)
    {
        # $milter_conn undefined, so we probably need to grep
        # sendmail.cf for it... or just panic.
        $milter_conn = auto_getconn($milter_name, $milter_cf);
        print STDERR "Got Milter config: $milter_conn\n" if ($verbose);
    }
    die "No milter connection definition found!\n" unless ($milter_conn);

    my $sleep_count = 0;
    do {
        $milter_pid = fork();
        unless (defined $milter_pid) {
            warn "cannot fork: $!";
            die "bailing out" if $sleep_count++ > 6;
            sleep 2 + $sleep_count;
        }
    } until defined $milter_pid;
    if ($milter_pid > 0)
    {
        $SIG{INT}  = sub { kill 15, $milter_pid; exit(0); };
        $SIG{TERM} = sub { kill 15, $milter_pid; exit(0); };
        return;
    }

    # Am kid, start the milter and then restart it if it dies.
    while (1)
    {
        print STDERR "Starting F-Prot Milter...\n" if ($verbose);
        system $prog ($prog, "-p", $milter_conn);
        sleep(1);
    }
}

sub scan_job
{
    my ($job) = @_;

    # Unescape HTTP-style encoding of odd characters
    $job =~ s/[\n\r]+$//;
    my ($ijob, $ojob) = split(/\t/, $job);
    my $rjob = $ijob;
    $ijob =~ s/%([0-9a-fA-F][0-9a-fA-F])/chr(hex($1))/gse;
    $ojob =~ s/%([0-9a-fA-F][0-9a-fA-F])/chr(hex($1))/gse;
    $ojob = $ijob.".out" unless ($ojob);

    return scan_jobs($ijob, $ojob, $rjob);
}

sub milter_scan_jobs
{
    my $sleep_count = 0;
    my $pid = undef;
    do {
        $pid = open(SCAN_KID, "-|");
        unless (defined $pid) {
            warn "cannot fork: $!";
            die "bailing out" if $sleep_count++ > 6;
            sleep 2 + $sleep_count;
        }
    } until defined $pid;
    
    if ($pid > 0)
    {
        my @stat = split(/:\s+/, join('', <SCAN_KID>), 4);
        close(SCAN_KID);
        return @stat;
    }
    else
    {
        print STDOUT scan_jobs(@_);
        exit(0);
    }
}

sub scan_jobs
{
    my ($ijob, $ojob, $rjob) = @_;
    $rjob = $ijob unless ($rjob);

    # Reopen STDIN/STDOUT so we can use the unmodified do_stat().
    print STDERR "SCANNING: $ijob -> $ojob\n" if ($verbose);
    open(STDOUT_OLD, ">&STDOUT");
    close(STDIN);
    close(STDOUT);
    open(STDIN, "<$ijob") || return "$rjob: 200: 0: Couldn't open $ijob, $!";
    open(STDOUT,">$ojob") || return "$rjob: 201: 0: Couldn't create $ojob, $!";

    my ($code, $modifications, $text) = do_scan();

    close(STDOUT);
    close(STDIN);
    open(STDOUT, ">&STDOUT_OLD");
    close(STDOUT_OLD);
    
    my $ret = "$rjob: $code: $modifications: $text";
    print STDERR $ret, "\n" if ($verbose);
    return $ret;
}

sub do_scan
{
    my $modifications = 0;
    # Time to scan.. backup if the user wants..
    if ($backup || $quarantine)
    {
        my $time = time();
        my (undef, undef, undef, $d, $m, undef, undef, undef) = localtime($time);
        $d = sprintf("%02d", $d);
        $m = sprintf("%02d", $m+1);
        my $subdir = "/".$m.$d."/";
        unless (-d $backup_location)
        {
            unless (mkdir($backup_location,1777))
            {
                return (100, 0,
                        "Unable to create backup basedir $backup_location: $!\n");
            }
            chmod oct(1777), $backup_location;
        }
        my $pathto = $backup_location.$subdir;
        
        unless(-d $pathto)
        {
              unless (mkdir($pathto,1777))
            {
                return (101, 0,
                        "Unable to create backup subdir $pathto: $!\n");
            }
            chmod oct(1777), $pathto;
        }
        my $backup_filename = $pathto.$time.".".$$;

        unless (open(BACKUP_FILE, "+> ".$backup_filename))
        {
            return (102, 0,
                    "Cannot open backup file $backup_filename: $!\n");
        }
        
        while (<STDIN>)
        {
            print BACKUP_FILE $_;
        }
        seek(BACKUP_FILE, 0, 0);
        $ret = $engine->sanitize(*BACKUP_FILE, *STDOUT);
        $modifications = ($engine->{"mod_id"} - $engine->{"base_mod_id"});
        close(BACKUP_FILE);
        
        unless ($modifications || $backup)
        {
            # Message wasn't modified and backups weren't requested, 
            # so delete from quarantine.
            unlink($backup_filename);
        }

        # Try to delete the directory. Will fail unless it's empty.
        rmdir($pathto);
    }
    else
    {
        $ret = $engine->sanitize(*STDIN, *STDOUT);
        $modifications = ($engine->{"mod_id"} - $engine->{"base_mod_id"});
    }
    
    if (my $e = $engine->error())
    {
        print STDERR $e;
    }

    my @messages = ( );
    if (@fprot_codes)
    {
        push @messages, "Error(s) scanning: ".join(', ', @fprot_codes);
        $ret += 10;
    }
    if (@infections)
    {
        push @messages, "Detected: ".join(', ', @infections);
    }

    return ($ret, $modifications, 
            join(', ', "Mod score: $modifications", @messages));
}

sub auto_getconn 
{
    my ($milter, $cf) = @_;
    local *CF;
    
    open(CF, '<'.$cf) or die "open($cf): $!";
    while (<CF>) 
    {
        s/\s+$//; # also trims newlines
        
        next unless (s/^X([^,\s]+),\s*//);
        next unless ($milter eq $1);

        # We've found our milter definition, now "parse" it.
        while (s/^(.)=([^,\s]+)(,\s*|\Z)//) 
        {
            if ($1 eq 'S')
            {
                close(CF);
                return $2;
            }
        }
    }
    
    close(CF);
    undef;
}
